// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
// some code from https://01.org/linuxgraphics/gfx-docs/drm/media/uapi/v4l/v4l2grab.c.html
#include <fcntl.h>
#include <libv4l2.h>
#include <string.h>
#include <unistd.h>
#include <linux/videodev2.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include "error.h"
#include "source.h"
#include "source_v4l.h"
#include "picio.h"
#include "log.h"
#include "utils.h"

static void xioctl(int fh, long unsigned int request, void *arg)
{
        int r;

        do {
                r = v4l2_ioctl(fh, request, arg);
        } while (r == -1 && (errno == EINTR || errno == EAGAIN));

        if (r == -1) {
                fprintf(stderr, "error %d, %s\n", errno, strerror(errno));
                exit(EXIT_FAILURE);
        }
}

source_v4l::source_v4l(const std::string & id, const std::string & descr, const std::string & exec_failure, const std::string & dev, const std::string & pixel_format, const bool rpi_workaround_in, const int jpeg_quality, const double max_fps, const int w_override, const int h_override, resize *const r, const int resize_w, const int resize_h, const int loglevel, const double timeout, std::vector<filter *> *const filters) : source(id, descr, exec_failure, max_fps, r, resize_w, resize_h, loglevel, timeout, filters), pixel_format(pixel_format), rpi_workaround(rpi_workaround_in)
{
	fd = v4l2_open(dev.c_str(), O_RDWR, 0);
	if (fd == -1)
		error_exit(true, "Cannot access video4linux device '%s'", dev.c_str());

	struct v4l2_format              fmt;
        struct v4l2_buffer              buf;
        struct v4l2_requestbuffers      req;
        enum v4l2_buf_type              type;

	memset(&fmt, 0x00, sizeof fmt);
        fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = w_override;
        fmt.fmt.pix.height      = h_override;
        fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
        fmt.fmt.pix.field       = V4L2_FIELD_INTERLACED;
	//printf("VIDIOC_S_FMT\n");
        xioctl(fd, VIDIOC_S_FMT, &fmt);
        if (fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_RGB24)
                error_exit(false, "Libv4l didn't accept RGB24 format. Can't proceed.");

        if (fmt.fmt.pix.width != w_override || fmt.fmt.pix.height != h_override)
                log(LL_INFO, "Warning: driver is sending image at %dx%d (requested: %dx%d)", fmt.fmt.pix.width, fmt.fmt.pix.height, w_override, h_override);
	width = fmt.fmt.pix.width;
	height = fmt.fmt.pix.height;

	pixelformat = fmt.fmt.pix.pixelformat;

	memset(&req, 0x00, sizeof req);
        req.count = 2;
        req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_MMAP;
	// printf("VIDIOC_REQBUFS\n");
        xioctl(fd, VIDIOC_REQBUFS, &req);
	// printf("got: %d\n", req.count);

	buffers = (buffer *)calloc(req.count, sizeof(*buffers));
	n_buffers = 0;
        for(; n_buffers < req.count; ++n_buffers) {
                memset(&buf, 0x00, sizeof(buf));
                buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory      = V4L2_MEMORY_MMAP;
                buf.index       = n_buffers;

		// printf("VIDIOC_QUERYBUF\n");
                xioctl(fd, VIDIOC_QUERYBUF, &buf);

                buffers[n_buffers].length = buf.length;
                buffers[n_buffers].start = v4l2_mmap(NULL, buf.length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, buf.m.offset);

                if (MAP_FAILED == buffers[n_buffers].start) {
                        perror("mmap");
                        exit(EXIT_FAILURE);
                }
        }

	log(LL_INFO, "v4l: n_buffers: %d", n_buffers);
        for (int i = 0; i < n_buffers; ++i) {
                memset(&buf, 0x00, sizeof(buf));
                buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                buf.memory = V4L2_MEMORY_MMAP;
                buf.index = i;
                //printf("VIDIOC_QBUF\n");
                xioctl(fd, VIDIOC_QBUF, &buf);
        }

        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        //printf("VIDIOC_STREAMON\n");
        xioctl(fd, VIDIOC_STREAMON, &type);

	vw = width;
	vh = height;

	if (need_scale()) {
		width = resize_w;
		height = resize_h;
	}

	th = NULL;
}

source_v4l::~source_v4l()
{
	stop();

	enum v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        xioctl(fd, VIDIOC_STREAMOFF, &type);

        for (int i = 0; i < n_buffers; ++i)
                v4l2_munmap(buffers[i].start, buffers[i].length);

	v4l2_close(fd);

	free(buffers);

}

void source_v4l::operator()()
{
	log(id, LL_INFO, "source v4l2 thread started");

	set_thread_name("src_v4l2");

	int bytes = vw * vh * 3;
	unsigned char *conv_buffer = static_cast<unsigned char *>(allocate_aligned(bytes));

	const uint64_t interval = max_fps > 0.0 ? 1.0 / max_fps * 1000.0 * 1000.0 : 0;

	for(;!local_stop_flag;)
	{
		struct v4l2_buffer buf;
		memset(&buf, 0x00, sizeof buf);
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;

		uint64_t start_ts = get_us();

		if (v4l2_ioctl(fd, VIDIOC_DQBUF, &buf) == -1) {
			log(id, LL_ERR, "VIDIOC_DQBUF failed: %s", strerror(errno));
		}
		else if (work_required() && !is_paused()) {
			unsigned char *io_buffer = (unsigned char *)buffers[buf.index].start;

			if (pixelformat == V4L2_PIX_FMT_JPEG) {
				int cur_n_bytes = buf.bytesused;

				if (resize_h != -1 || resize_w != -1) {
					int dw = -1, dh = -1;
					unsigned char *temp = NULL;
					if (read_JPEG_memory(io_buffer, cur_n_bytes, &dw, &dh, &temp))
						set_scaled_frame(temp, dw, dh);
					free(temp);
				}
				else {
					set_frame(E_JPEG, io_buffer, cur_n_bytes);
				}
			}
			else {
				memcpy(conv_buffer, io_buffer, vw * vh * 3);

				if (need_scale())
					set_scaled_frame(conv_buffer, vw, vh);
				else
					set_frame(E_RGB, conv_buffer, vw * vh * 3);
			}
		}

		if (v4l2_ioctl(fd, VIDIOC_QBUF, &buf) == -1)
			error_exit(true, "ioctl(VIDIOC_QBUF) failed");

		uint64_t end_ts = get_us();
		int64_t left = interval - (end_ts - start_ts);

		//printf("total took %d, left %d, interval %ld\n", int(end_ts - start_ts), int(left), int(interval));

		if (interval > 0 && left > 0)
			myusleep(left);
	}

	free(conv_buffer);

	log(id, LL_INFO, "source v4l2 thread terminating");
}
